""" Module to manage EVA robot operations with RobotManager """
import logging
import time

from .robot_api import RobotManagerHTTP, RobotManagerSimulator


class RobotException(Exception):
    pass


class Robot:
    """ Main Robot class to interface with a RobotManager instance
        :param robot_name: name of the current pipetting robot;
        :param robot_manager_host: hostname or ip address of the RobotManager instance;
        :param simulate: whether to simulate action requests and check results;
        :param check_wait_time: delay between *check_action* calls waiting for the action to be completed;
    """
    def __init__(self,
                 robot_name: str,
                 robot_manager_host: str,
                 robot_manager_port: int,
                 simulate: bool = False,
                 check_wait_time: float = 0.5):
        if not robot_name.isalnum():
            raise RobotException("Robot name not aphanumeric: {}".format(robot_name))
        self._robot_name = robot_name
        self._logger = logging.getLogger(__name__)
        self._logger.info("Simulate is {}".format(simulate))
        self._check_wait_time = 0 if simulate else check_wait_time
        self._api = RobotManagerSimulator() if simulate else RobotManagerHTTP(robot_manager_host, robot_manager_port)

    def build_request(self, action: str, slot: str, plate_name: str, ):
        return {
            "action": action,
            "machine": self._robot_name,
            "position": "{}-{}".format(self._robot_name, slot),
            "plate_name": plate_name
        }

    def pick_plate(self, slot: str, plate_name: str):
        self.execute_action("pick", slot, plate_name)

    def drop_plate(self, slot: str, plate_name: str):
        self.execute_action("drop", slot, plate_name)

    def execute_action(self, action, slot, plate_name):
        action_id = self._api.action_request(self.build_request(action, slot, plate_name))
        self.wait_for_action_to_finish(action_id)

    def wait_for_action_to_finish(self, action_id):
        self._logger.info("Waiting for action to finish with id: {}".format(action_id))
        while True:
            res = self._api.check_action(action_id)
            self._logger.info("Received {}".format(res))
            if res["status"] != "pending":
                break
            else:
                if self._check_wait_time:
                    time.sleep(self._check_wait_time)



""" Base class that instantiate robot control """
from abc import ABC
from covmatic_stations.station import Station, instrument_loader, labware_loader
from .robot import Robot


class RobotStationABC(Station, ABC):
    def __init__(self,
                 ot_name: str,
                 robot_manager_host: str = None,
                 robot_manager_port: int = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._ot_name = ot_name
        self._robot_manager_host = robot_manager_host
        self._robot_manager_port = robot_manager_port

    @instrument_loader(0, "_robot")
    def load_robot(self):
        self._robot = Robot(robot_name=self._ot_name,
                            robot_manager_host=self._robot_manager_host,
                            robot_manager_port=self._robot_manager_port,
                            simulate=self._ctx.is_simulating())

    def robot_pick_plate(self, slot, plate_name):
        self.home()
        self._robot.pick_plate(slot, plate_name)

    def robot_drop_plate(self, slot, plate_name):
        self.home()
        self._robot.drop_plate(slot, plate_name)



import unittest
from unittest.mock import patch, MagicMock

from src.covmatic_robotstation.robot_api import RobotManagerHTTP, RobotManagerHTTPException

FAKE_ACTION_ID = "fakeactionid"
FAKE_HOST = "HOST"
FAKE_PORT = 1234

ACTION_PICK = {
            "action": "pick",
            "machine": "OT1",
            "position": "SLOT1",
            "plate_name": "REAGENT"
        }
ACTION_EXPECTED_URL = "http://{host}:{port}/action/pick/OT1/SLOT1/REAGENT".format(host=FAKE_HOST, port=FAKE_PORT)
ACTION_ANSWER = {
    "action_id": FAKE_ACTION_ID
}
MALFORMED_ACTION_ANSWER = {
    "wrong": "field"
}

CHECK_EXPECTED_URL = "http://{host}:{port}/action/check/{id}".format(host=FAKE_HOST, port=FAKE_PORT, id=FAKE_ACTION_ID)
CHECK_EXPECTED_ANSWER = {
    "state": "finished"
}
CHECK_MALFORMED_ANSWER = {
    "some": "thing"
}
CHECK_DATA_RETURNED = {
    "state": "finished"
}


class TestAPI(unittest.TestCase):
    def setUp(self) -> None:
        self._requests_patcher = patch('src.covmatic_robotstation.robot_api.requests')
        self._mock_requests = self._requests_patcher.start()
        self._mock_response = MagicMock()
        self._mock_requests.get.return_value = self._mock_response
        self._api = RobotManagerHTTP(FAKE_HOST, FAKE_PORT)

    def tearDown(self) -> None:
        self._requests_patcher.stop()


class TestActionRequest(TestAPI):
    def setUp(self) -> None:
        super().setUp()
        self._mock_response.json.return_value = ACTION_ANSWER
        self._mock_response.status_code = 200

    def test_request_calls_get(self):
        self._api.action_request(ACTION_PICK)
        self._mock_requests.get.assert_called_once()

    def test_request_return_value(self):
        self.assertEqual(FAKE_ACTION_ID, self._api.action_request(ACTION_PICK))

    def test_request_get_404(self):
        self._mock_response.json.side_effect = Exception("We shouldn't look for this json answer")
        self._mock_response.status_code = 404

        with self.assertRaises(RobotManagerHTTPException):
            self._api.action_request(ACTION_PICK)
        self._mock_requests.get.assert_called_once()

    def test_wrong_answer_raises(self):
        self._mock_response.json.return_value = MALFORMED_ACTION_ANSWER
        with self.assertRaises(RobotManagerHTTPException):
            self._api.action_request(ACTION_PICK)

    def test_request_url(self):
        self._api.action_request(ACTION_PICK)
        self._mock_requests.get.assert_called_once_with(ACTION_EXPECTED_URL)


class TestCheckOk(TestAPI):
    def setUp(self) -> None:
        super().setUp()
        self._mock_response.json.return_value = CHECK_EXPECTED_ANSWER
        self._mock_response.status_code = 200

    def test_calls_get(self):
        self._api.check_action(FAKE_ACTION_ID)
        self._mock_requests.get.assert_called_once()

    def test_called_url(self):
        self._api.check_action(FAKE_ACTION_ID)
        self._mock_requests.get.assert_called_once_with(CHECK_EXPECTED_URL)

    def test_return_value(self):
        # self._mock_requests.get.side_effect = [CHECK_EXPECTED_ANSWER]
        self.assertEqual(CHECK_DATA_RETURNED, self._api.check_action(FAKE_ACTION_ID))

    def test_calls_get_404(self):
        self._mock_response.json.side_effect = Exception("We shouldn't look for this json answer")
        self._mock_response.status_code = 404

        with self.assertRaises(RobotManagerHTTPException):
            self._api.check_action(FAKE_ACTION_ID)
        self._mock_requests.get.assert_called_once()

    def test_calls_get_wrong_answer(self):
        self._mock_response.json.side_effect = [CHECK_MALFORMED_ANSWER]

        with self.assertRaises(RobotManagerHTTPException):
            self._api.check_action(FAKE_ACTION_ID)
        self._mock_requests.get.assert_called_once()



import unittest
from unittest.mock import patch

from src.covmatic_robotstation.robot import Robot, RobotException

FAKE_ROBOT_MANAGER_HOST = "fakerobotmanagerhost"
FAKE_ROBOT_MANAGER_PORT = 1234

OT_NAME = "OT"
WRONG_NAME = "OT 1"

SLOT_NAME = "SLOT1"
POSITION = "{}-{}".format(OT_NAME, SLOT_NAME)

PLATE_NAME = "PLATE1"

expected_pick_action = {
            "action": "pick",
            "machine": OT_NAME,
            "position": POSITION,
            "plate_name": PLATE_NAME
        }

expected_drop_action = {
            "action": "drop",
            "machine": OT_NAME,
            "position": POSITION,
            "plate_name": PLATE_NAME
        }

FAKE_ID_PICK = "FAKEIDPICK"
FAKE_ID_DROP = "FAKEIDDROP"

PENDING_STATUS = "pending"
FINISHED_STATUS = "finished"

CHECK_RETURN_VALUE_PENDING = {
    "status": PENDING_STATUS
}

CHECK_RETURN_VALUE_FINISHED = {
    "status": FINISHED_STATUS
}


class TestRobot(unittest.TestCase):
    def setup_mocks(self):
        self._api_patcher = patch('src.covmatic_robotstation.robot.RobotManagerHTTP')
        self._sleep_patcher = patch('src.covmatic_robotstation.robot.time.sleep')
        self._mock_api = self._api_patcher.start()
        self._mock_sleep = self._sleep_patcher.start()

    def setUp(self):
        self.setup_mocks()
        self._r = Robot(OT_NAME, FAKE_ROBOT_MANAGER_HOST, FAKE_ROBOT_MANAGER_PORT)

    def tearDown(self) -> None:
        self._api_patcher.stop()
        self._sleep_patcher.stop()


class TestWrongName(TestRobot):
    def setUp(self):
        self.setup_mocks()

    def test_name_not_alphanum(self):
        with self.assertRaises(RobotException):
            Robot(WRONG_NAME, FAKE_ROBOT_MANAGER_HOST, FAKE_ROBOT_MANAGER_PORT)


class TestPickFunction(TestRobot):
    def setUp(self):
        super().setUp()
        self._r.pick_plate(SLOT_NAME, PLATE_NAME)

    def test_request(self):
        self._mock_api().action_request.assert_called_once()

    def test_request_argument(self):
        self._mock_api().action_request.assert_called_once_with(expected_pick_action)


class TestPickFunctionWaitForComplete(TestRobot):
    def setUp(self):
        super().setUp()
        self._mock_api().action_request.side_effect = [FAKE_ID_PICK]

    def test_request_calls_check(self):
        self._r.pick_plate(SLOT_NAME, PLATE_NAME)
        self._mock_api().check_action.assert_called()

    def test_request_calls_check_with_action_id(self):
        self._r.pick_plate(SLOT_NAME, PLATE_NAME)
        self._mock_api().check_action.assert_called_with(FAKE_ID_PICK)

    def test_request_read_check_return_value(self):
        self._mock_api().check_action.side_effect = [CHECK_RETURN_VALUE_FINISHED]

        self._r.pick_plate(SLOT_NAME, PLATE_NAME)

        self._mock_api().check_action.assert_called_once_with(FAKE_ID_PICK)

    def test_request_keep_calling_while_pending(self):
        self._mock_api().check_action.side_effect = [CHECK_RETURN_VALUE_PENDING, CHECK_RETURN_VALUE_FINISHED]

        self._r.pick_plate(SLOT_NAME, PLATE_NAME)

        self._mock_api().check_action.assert_called_with(FAKE_ID_PICK)
        self.assertEqual(self._mock_api().check_action.call_count, 2)

    def test_request_uses_timed_request(self):
        self._mock_api().check_action.side_effect = [CHECK_RETURN_VALUE_PENDING, CHECK_RETURN_VALUE_FINISHED]

        self._r.pick_plate(SLOT_NAME, PLATE_NAME)

        self._mock_sleep.assert_called()

    def test_request_keep_calling_while_pending_long(self):
        self._mock_api().check_action.side_effect = [CHECK_RETURN_VALUE_PENDING for _ in range(99)] + [CHECK_RETURN_VALUE_FINISHED]

        self._r.pick_plate(SLOT_NAME, PLATE_NAME)

        self._mock_api().check_action.assert_called_with(FAKE_ID_PICK)
        self.assertEqual(self._mock_api().check_action.call_count, 100)


class TestDropFunction(TestRobot):
    def setUp(self):
        super().setUp()
        self._r.drop_plate(SLOT_NAME, PLATE_NAME)

    def test_request(self):
        self._mock_api().action_request.assert_called_once()

    def test_request_argument(self):
        self._mock_api().action_request.assert_called_once_with(expected_drop_action)


class TestDropFunctionWaitForComplete(TestRobot):
    def setUp(self):
        super().setUp()
        self._mock_api().action_request.side_effect = [FAKE_ID_DROP]

    def test_request_calls_check(self):
        self._r.drop_plate(SLOT_NAME, PLATE_NAME)
        self._mock_api().check_action.assert_called()

    def test_request_calls_check_with_action_id(self):
        self._r.drop_plate(SLOT_NAME, PLATE_NAME)
        self._mock_api().check_action.assert_called_with(FAKE_ID_DROP)

    def test_request_read_check_return_value(self):
        self._mock_api().check_action.side_effect = [CHECK_RETURN_VALUE_FINISHED]

        self._r.drop_plate(SLOT_NAME, PLATE_NAME)

        self._mock_api().check_action.assert_called_once_with(FAKE_ID_DROP)

    def test_request_keep_calling_while_pending(self):
        self._mock_api().check_action.side_effect = [CHECK_RETURN_VALUE_PENDING, CHECK_RETURN_VALUE_FINISHED]

        self._r.drop_plate(SLOT_NAME, PLATE_NAME)

        self._mock_api().check_action.assert_called_with(FAKE_ID_DROP)
        self.assertEqual(self._mock_api().check_action.call_count, 2)

    def test_request_uses_timed_request(self):
        self._mock_api().check_action.side_effect = [CHECK_RETURN_VALUE_PENDING, CHECK_RETURN_VALUE_FINISHED]

        self._r.drop_plate(SLOT_NAME, PLATE_NAME)

        self._mock_sleep.assert_called()

    def test_request_keep_calling_while_pending_long(self):
        self._mock_api().check_action.side_effect = [CHECK_RETURN_VALUE_PENDING for _ in range(99)] + [CHECK_RETURN_VALUE_FINISHED]

        self._r.drop_plate(SLOT_NAME, PLATE_NAME)

        self._mock_api().check_action.assert_called_with(FAKE_ID_DROP)
        self.assertEqual(self._mock_api().check_action.call_count, 100)



from covmatic_robotstation.robot_station import RobotStationABC, labware_loader, instrument_loader

class FakeStation(RobotStationABC):
    @labware_loader(0, "_fake_tipracks")
    def load_fake_tiprack(self):
        self._fake_tipracks = [self._ctx.load_labware('opentrons_96_filtertiprack_20ul', "1", '20ul filter tiprack')]

    @instrument_loader(0, "_fake_pipette")
    def load_fake_pipette(self):
        self._fake_pipette = self._ctx.load_instrument('p20_multi_gen2', 'left', tip_racks=self._fake_tipracks)

    def _tipracks(self) -> dict:
        return {"_fake_tipracks": "_fake_pipette"}

    def body(self):
        self.robot_pick_plate("SLOT1", "TESTPLATE")
        self.robot_drop_plate("SLOT2", "TESTPLATE")


metadata = {'apiLevel': '2.7'}
station = FakeStation(ot_name="OTTEST", num_samples=96)


def run(ctx):
    return station.run(ctx)


if __name__ == "__main__":
    FakeStation(ot_name="OT1", metadata={'apiLevel': '2.7'}).simulate()



""" API to interface with RobotManager API """
import logging
from abc import ABC, abstractmethod
import requests
import logging


class RobotManagerInterface(ABC):
    """ Common interface for the API towards RobotManager"""
    @abstractmethod
    def action_request(self, action_dict) -> str:
        pass

    @abstractmethod
    def check_action(self, action_id) -> dict:
        pass


class RobotManagerHTTPException(Exception):
    pass


class RobotManagerHTTP(RobotManagerInterface):
    def __init__(self, host: str, port: int, logger=None):
        self._host = host
        self._port = port
        self._logger = logger or logging.getLogger(self.__class__.__name__)
        self._logger.info("Starting with host {}".format(self._host))

    def action_request(self, action_dict) -> str:
        self._logger.info("Requesting action {}".format(action_dict))
        answer = requests.get(self._get_url_from_action(action_dict))
        self._logger.info("Received answer: {}".format(answer))
        if answer.status_code == 200:
            answer_json = answer.json()
            if "action_id" in answer_json:
                return answer_json["action_id"]
            raise RobotManagerHTTPException("Unexpected answer to action request: {}".format(answer))
        raise RobotManagerHTTPException("Action request {} http code not good: {}".format(action_dict, answer))

    def _get_url_from_action(self, action_dict):
        _url = "{baseurl}/action/{action}/{machine}/{position}/{plate_name}".format(baseurl=self._baseurl, **action_dict)
        self._logger.info("Returning url {} for action {}".format(_url, action_dict))
        return _url

    def check_action(self, action_id) -> dict:
        _url = "{baseurl}/action/check/{action_id}".format(baseurl=self._baseurl, action_id=action_id)
        answer = requests.get(_url)
        if answer.status_code == 200:
            answer_json = answer.json()
            if "state" in answer_json:
                return answer_json
            raise RobotManagerHTTPException("Check unexpected answer for id {}: {}".format(action_id, answer))
        raise RobotManagerHTTPException("Check id {} has error: {}".format(action_id, answer))

    @property
    def _baseurl(self) -> str:
        return "http://{host}:{port}".format(host=self._host, port=self._port)

class RobotManagerSimulator(RobotManagerInterface):
    """ Simulation API """
    def __init__(self, *args, **kwargs):
        self._check_count = 0
        self._check_before_finish = 10
        self._logger = logging.getLogger(__name__)
        self._logger.info("{} starting!".format(self.__class__.__name__))
        self._logger.info("Received args: {} and {}".format(args, kwargs))

    def action_request(self, action_dict) -> str:
        self._logger.info("Received action request: {}".format(action_dict))
        return "fakeactionid"

    def check_action(self, action_id) -> dict:
        self._logger.info("Received check request for action id: {}".format(action_id))
        status = "pending"

        self._check_count += 1
        if self._check_count == self._check_before_finish:
            self._check_count = 0
            status = "finished"

        self._logger.info("Returning status {}".format(status))
        return {"status": status}

